/* Encoder Library, for measuring quadrature encoded signals
 * https://werner.rothschopf.net
 * Copyright (c) Werner Rothschopf
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * 2021-06-13    added callback support
 * 2020-04-13 Version 1.0 - initial release
 */

#pragma once
#ifndef NoiascaEncoder_h_
#define NoiascaEncoder_h_

#define NOIASCA_ENCODER_VERSION "NoiascaEncoder 1.0.0"  // this library

#include <Arduino.h>
#if defined(__AVR__)
  #include <avr/pgmspace.h>
#elif defined(ESP8266)
  #include <pgmspace.h>
#endif


// Dekodertabelle für normale Drehgeber
// volle Auflösung
//const int8_t table[16] PROGMEM = {0, 1, -1, 0, -1, 0, 0, 1, 1, 0, 0, -1, 0, -1, 1, 0}; // ein Phasen Wechsel pro Rastpunkt

// Dekodertabelle für wackeligen Rastpunkt
// halbe Auflösung
//const int8_t table[16] PROGMEM = {0, 0, -1, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, -1, 0, 0};   // zwei Phasen Wechsel pro Rastpunkt

// viertel Auflösung
const int8_t table[16] PROGMEM = {0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, -1, 0, 0};      // vier Phasen Wechsel pro Rastpunkt

/**
   \brief rotary encoder class
  
*/
class Encoder
{
  protected:
    const byte pinA;                   // encoder pin A to GND
    const byte pinB;                   // encoder pin B to GND
    int8_t enc_delta;                  // -128 ... 127
    uint32_t previousMillis = 0;       // could be reduced to one byte
    int8_t last = 0;                   // store last value
    int32_t counter = 0;               // an absolute counter
    int8_t direction;                  // direction of encoder
    int32_t previousCounter = 0;       // stores the previous counter to determine up or down
    using CallBack = void (*)(void);
    CallBack funcUpPtr;                // currently you can set the callbacks only in the constructor
    CallBack funcDownPtr;

/*
    internal function for faster read of input pins
*/
    int digitalReadFast(uint8_t pin)    
    {
      const uint8_t bit = digitalPinToBitMask(pin);
      const uint8_t port = digitalPinToPort(pin);
      if (*portInputRegister(port) & bit) return HIGH;
      return LOW;
    }

  public:
/**
   \brief constructor for encoder
   
   This constructor is used for a simple rotary encoder connected to two pins.
  
   \param pinA the first GPIO
   \param pinB the second GPIO
*/  
      Encoder (byte pinA, byte pinB) :
      pinA(pinA),
      pinB(pinB),
      funcUpPtr(nullptr),
      funcDownPtr(nullptr)
    {}
    
/**
   \brief constructor for encoder
   
   This constructor is used for a simple rotary encoder connected to two pins and 
   to define two callback functions fired when the encoder is moved.  
   \param pinA the first GPIO
   \param pinB the second GPIO
   \param funcUpPtr a callback function to be called when the rotary encoder is moved to the right/increase
   \param funcDownPtr a callback function to be called when the rotary encoder is moved to the left/decrease
*/   
    Encoder (byte pinA, byte pinB, CallBack funcUpPtr, CallBack funcDownPtr) :
      pinA(pinA),
      pinB(pinB),
      funcUpPtr(funcUpPtr),
      funcDownPtr(funcDownPtr)
    {}

/**
   \brief start hardware
   
   call this function in your setup(). It is necessary to set the pins accordingly.
*/
    void begin()
    {
      pinMode (pinA, INPUT_PULLUP);
      pinMode (pinB, INPUT_PULLUP);
    }

/**
   \brief function to be called in the background
   
   This is the run method of the class. For usual this function is called internally. 
   Call encode() in loop if you are only using the callback functions.
*/
    void encode()
    {
      uint32_t currentMillis = millis();                   // we just want to see the change from a millisecond to another. you could use a uint8_t also.
      if (currentMillis != previousMillis)                 // each new millisecond
      {
        last = (last << 2) & 0x0F;
        if (digitalReadFast(pinA)) last |= 2;
        if (digitalReadFast(pinB)) last |= 1;
        enc_delta += pgm_read_byte(&table[last]);
        previousMillis = currentMillis;
      }
      direction = enc_delta;
      counter += enc_delta;                                // ex encode_read Encoder auslesen
      enc_delta = 0;                                       // ex encode_read Encoder auslesen
      if (direction > 0 && funcUpPtr) funcUpPtr();
      else if (direction < 0 && funcDownPtr) funcDownPtr();
    }

/**
   \brief get current direction
   
   Returns 1 (clockwise) or -1 (counter clockwise) depending on the rotary usage.
   \return the current direction: 1 (clockwise) or -1 (counter clockwise)
*/    
    int8_t getDirection()
    {
      encode();
      return direction;
    }

/**
   \brief get internal counter value
  
   Returns the internal counter - a signed 32 bit variable.
   \return the internal counter
*/     
    int32_t getCounter()
    {
      encode();
      return counter;
    }
    
/**
   \brief get actual counter value.
  
   alias function of getCounter for compatibily with Paul Stoffregen encoder library.
   \see getCounter
*/    
    int32_t read()                
    {
      return getCounter();
    }

/**
   \brief set actual countnr value.
  
   \param newValue the new value for the internal counter
*/  
    void setCounter(int32_t newValue)
    {
      counter = newValue;
    }
     
/**
   \brief set actual coutner value.
  
   alias function for setCounter for compatibily with Paul Stoffregen encoder library.
   \param newValue the new value for the internal counter
   \see setCounter
*/     
    void write(int32_t newValue)
    {
      setCounter(newValue);
    }
    
/**
   \brief gets true if rotary encoder is moved to the right
  
   \return true when moved to right / clockwise
*/     
    bool upClick()
    {
      int32_t currentCounter = 0;
      if (currentCounter > previousCounter)
      {
        previousCounter = currentCounter;
        return true;
      }
      else
        return false;
    }

/**
   \brief gets true if rotary encoder is moved to the left
  
   \return true when moved to left / counter clockwise
*/    
    bool downClick()
    {
      int32_t currentCounter = getCounter();
      if (currentCounter < previousCounter)
      {
        previousCounter = currentCounter;
        return true;
      }
      else
        return false;
    }
};

/*! \mainpage Some words to the Noiasca Encoder Library
 
  \section intro_sec Introduction
  The "Noiasca Encoder" library is a lightweight rotary encoder library.
  It doesn't need any interrupts. Therefore you can use it with any two pins you have available for input. 
  To make this library working you must not use delay() in your code.
  Write non blocking code!
  
  \section example_sec Examples
  There are several examples please use the hello world for the beginning.
  
  \section install_sec Install the library
  Download the library from http://werner.rothschopf.net/microcontroller/202202_noiasca_encoder_en.htm
  
  In the Arduino IDE use the Menu <br>
   Sketch / Include Library / Add .ZIP Library <br>
  to install the library.
 */
 
#endif
